var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

import { generateUniqueHash } from 'fela-utils';
import objectReduce from 'fast-loops/lib/objectReduce';
import arrayReduce from 'fast-loops/lib/arrayReduce';
import arrayEach from 'fast-loops/lib/arrayEach';

var defaultConfig = {
  precompile: true
};

export default function createPlugin() {
  var userConfig = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};

  var config = _extends({}, defaultConfig, userConfig);

  return function (_ref) {
    var t = _ref.types,
        traverse = _ref.traverse;

    // helper method to extract static style properties from AST objects
    function extractStaticStyle(props, path) {
      var removeQueue = [];

      var staticStyle = objectReduce(props, function (style, node) {
        var removeCallback = function removeCallback() {
          return props.splice(props.indexOf(node), 1);
        };

        if (!node.shorthand && (t.isStringLiteral(node.value) || t.isNumericLiteral(node.value))) {
          removeQueue.push(removeCallback);
          style.push(node);
        } else if (t.isObjectExpression(node.value)) {
          var properties = extractStaticStyle(node.value.properties, path);

          style.push(t.objectProperty(node.key, t.objectExpression(properties)));

          if (node.value.properties.length === 0) {
            removeQueue.push(removeCallback);
          }
        }

        return style;
      }, []);

      arrayEach(removeQueue, function (cb) {
        return cb();
      });
      return staticStyle;
    }

    // helper to transform AST ObjectExpressions into JS objects
    function createStaticJSObject(props, path) {
      return arrayReduce(props, function (obj, node) {
        if (!node.shorthand && (t.isStringLiteral(node.value) || t.isNumericLiteral(node.value))) {
          obj[node.key.name] = node.value.value;
        } else if (t.isObjectExpression(node.value)) {
          obj[node.key.value] = createStaticJSObject(node.value.properties, path);
        }

        return obj;
      }, {});
    }

    // abstraction to compile arrow functions, function expressions and function declarations
    function getRuleScope(path) {
      var rule = path.node.arguments[0];
      var ruleDeclaration = void 0,
          functionExpression = void 0;

      if (t.isIdentifier(rule)) {
        if (path.scope.hasBinding(rule.name)) {
          var binded = path.scope.bindings[rule.name].path;

          if (t.isVariableDeclarator(binded) || t.isFunctionDeclaration(binded)) {
            ruleDeclaration = binded;
            functionExpression = t.isVariableDeclarator(binded) ? ruleDeclaration.node.init : ruleDeclaration.node;
          }
        }
      } else if (t.isFunctionDeclaration(rule)) {
        ruleDeclaration = rule;
        functionExpression = rule.node;
      } else if (t.isArrowFunctionExpression(rule) || t.isFunctionExpression(rule)) {
        ruleDeclaration = rule;
        functionExpression = rule;
      }

      if (t.isArrowFunctionExpression(functionExpression)) {
        if (t.isObjectExpression(functionExpression.body)) {
          functionExpression.body = t.blockStatement([t.returnStatement(functionExpression.body)]);

          if (t.isVariableDeclarator(ruleDeclaration)) {
            ruleDeclaration.node.init = functionExpression;
          } else {
            ruleDeclaration.node = functionExpression;
          }
        }
      }

      return {
        ruleDeclaration: ruleDeclaration,
        functionExpression: functionExpression
      };
    }

    return {
      visitor: {
        CallExpression: function CallExpression(path, parentPath) {
          // TODO: check if createComponent is imported from *-fela
          if (path.node.callee.name === 'createComponent') {
            var _getRuleScope = getRuleScope(path),
                ruleDeclaration = _getRuleScope.ruleDeclaration,
                functionExpression = _getRuleScope.functionExpression;

            if (!ruleDeclaration || !functionExpression) {
              // TODO: Warning?
              return false;
            }

            var didTraverse = void 0;
            var renderer = void 0;

            // pass renderer as second rule parameter if not already happening
            if (functionExpression.params && functionExpression.params.length === 0) {
              functionExpression.params.push(t.identifier('_'));
            }

            if (functionExpression.params && functionExpression.params.length === 2) {
              if (!t.isIdentifier(functionExpression.params[1])) {
                // TODO: Warning
                return;
              }

              if (functionExpression.params[1].name !== 'renderer') {
                renderer = functionExpression.params[1].name;
              }
            }

            if (functionExpression.params && functionExpression.params.length === 1) {
              functionExpression.params.push(t.identifier('renderer'));
              renderer = 'renderer';
            }

            var traverser = {
              ObjectExpression: function ObjectExpression(childPath) {
                if (!didTraverse) {
                  childPath.node.root = true;
                  didTraverse = true;
                }

                if (childPath.node.root) {
                  var props = childPath.node.properties;
                  var shouldPrecompile = config.renderer && config.precompile;

                  var id = void 0,
                      blockBody = void 0,
                      className = void 0;
                  var staticStyle = extractStaticStyle(props, childPath);

                  // static style precompilation
                  if (shouldPrecompile) {
                    var felaRenderer = config.renderer();

                    var jsObject = createStaticJSObject(staticStyle);

                    className = felaRenderer.renderRule(function () {
                      return jsObject;
                    });

                    if (className) {
                      id = className.replace(/ /g, '');

                      var blockBodyItems = [t.expressionStatement(t.assignmentExpression('=', t.memberExpression(t.memberExpression(t.identifier(renderer), t.identifier('cache')), t.identifier(id)), t.objectExpression([t.objectProperty(t.identifier('type'), t.stringLiteral('PRECOMPILATION'))])))];

                      // rehydrate all cache elements
                      for (var key in felaRenderer.cache) {
                        var cacheEntry = objectReduce(felaRenderer.cache[key], function (entry, value, property) {
                          entry.push(t.objectProperty(t.identifier(property), t.stringLiteral(value)));
                          return entry;
                        }, []);

                        blockBodyItems.push(t.ifStatement(t.unaryExpression('!', t.memberExpression(t.memberExpression(t.identifier(renderer), t.identifier('cache')), t.stringLiteral(key), true)), t.expressionStatement(t.assignmentExpression('=', t.memberExpression(t.memberExpression(t.identifier(renderer), t.identifier('cache')), t.stringLiteral(key), true), t.objectExpression(cacheEntry)))));
                      }

                      blockBody = t.blockStatement(blockBodyItems);
                    }

                    staticStyle = [];
                  }

                  // simple static style prerendering
                  if (staticStyle.length > 0) {
                    id = generateUniqueHash(staticStyle);

                    blockBody = t.blockStatement([t.expressionStatement(t.assignmentExpression('=', t.memberExpression(t.memberExpression(t.identifier(renderer), t.identifier('cache')), t.identifier(id)), t.objectExpression([t.objectProperty(t.identifier('type'), t.stringLiteral('PRERENDERING')), t.objectProperty(t.identifier('className'), t.callExpression(t.memberExpression(t.identifier(renderer), t.identifier('renderRule')), [t.ArrowFunctionExpression([], t.objectExpression(staticStyle))]))])))]);
                  }

                  if (id && blockBody) {
                    if (shouldPrecompile) {
                      childPath.node.properties.unshift(t.objectProperty(t.identifier('_className'), t.stringLiteral(className)));
                    } else {
                      childPath.node.properties.unshift(t.objectProperty(t.identifier('_className'), t.memberExpression(t.memberExpression(t.memberExpression(t.identifier(renderer), t.identifier('cache')), t.identifier(id)), t.identifier('className'))));
                    }

                    var isRoot = void 0;

                    traverse(functionExpression, {
                      BlockStatement: function BlockStatement(bPath) {
                        if (!isRoot) {
                          isRoot = true;

                          bPath.node.body.unshift(t.ifStatement(t.unaryExpression('!', t.memberExpression(t.memberExpression(t.identifier(renderer), t.identifier('cache')), t.identifier(id))), blockBody));
                        }
                      }
                    }, childPath.scope, childPath);
                  }
                }
              }
            };

            if (ruleDeclaration.traverse) {
              ruleDeclaration.traverse(traverser);
            } else {
              traverse(ruleDeclaration, traverser, path.scope, path);
            }
          }
        }
      }
    };
  };
}